
#define SER_CLASS 'c'
#define SER_CLASSPTR 'p'
#define SER_STRING 's'
#define SER_LONG 'l'
#define SER_UNSIGNED_LONG 'u'
#define SER_SHORT 'o'
#define SER_UNSIGNED_SHORT 'n'
#define SER_BEGINCLASS 'b'
#define SER_ENDCLASS 'e'

/* format

-adressen,werte und laengen werden als hexadezimalzahlen mit 8 stellen codiert.
-namen haben immer 32 zeichen (rechts mit spaces aufgefuellt)
-codierstring:
 b<adresse im typstring>
 c<adresse im codierstring> | s<adresse im datenstring> | [luon]<wert>
 [...]
 effffffff

 adresse im codierstring=ffffffff falls zeiger=0
 adresse im datenstring=ffffffff falls zeiger=0
-typstring:
 <klassenname>
 <name>
 <name>
 [...]
 <leer>
-datenstring:
 <laenge><string><laenge><string>...

*/

struct serialize;

struct serialize_elem
{
  char name[33];
  char type;
  unsigned long offset;
  char* (*func)(char*);
  serialize* desc;

  serialize_elem* next;
  serialize_elem()
  {
    func=0;
    name[0]=0;
    next=0;
    offset=0;
    desc=0;
    return;
  };
  ~serialize_elem()
  {
    if (next) delete next;
    return;
  };
};

struct known_class
{
  char name[33];
  char addr[9];
  known_class* next;
  known_class()
  {
    name[0]=0;
    addr[0]=0;
    next=0;
    return;
  };
  ~known_class()
  {
    if (next) delete next;
  };
};

inline void sercpy(char* dst,char* src)
{
  unsigned long i;
  for (i=0;i!=32;i++)
    {
      if (!src || !src[i]) break;
      dst[i]=src[i];
    };
  while (i!=32) dst[i++]=' ';
  dst[i]=0;
  return;
}

struct serialize
{
  char tname[33];
  serialize_elem* elements;

  serialize()
  {
    elements=0;
    return;
  };
  void sername(char* name)
  {
    elements=0;
    sercpy(tname,name);
    return;
  };

  ~serialize()
  {
    delete elements;
  };

  inline void seradd(char* name,char type,unsigned long offset,serialize* desc)
  {
    serialize_elem* s=new serialize_elem();
    if (!s) log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
    sercpy(s->name,name);
    s->type=type;
    s->offset=offset; // -1;
    s->desc=desc;
    s->next=elements;
    elements=s;
    return;
  };
  inline void serfunc(char* name,char type,char* (*func)(char*),
		      serialize* desc)
  {
    serialize_elem* s=new serialize_elem();
    if (!s) log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
    sercpy(s->name,name);
    s->type=type;
    s->func=func;
    s->desc=desc;
    s->next=elements;
    elements=s;
    return;
  };

  inline void myrealloc(char*& zgr,unsigned long& len)
  {
    log(LOG_WARN,__FILE__,__LINE__,ERROR_UNSPECIFIED,
        "realloc from %lu to %lu",len,len*2);
    zgr=(char*) realloc(zgr,len*2);
    if (!zgr) log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
    len=len*2;
    return;
  };
  inline void sercode(char* obj,
		      char*& code,
		      unsigned long& pos,
		      char*& werte,
		      unsigned long& pos1,
		      char*& types,
		      unsigned long& pos2,
		      known_class*& known_classes,
		      unsigned long& maxlen_code,
		      unsigned long& maxlen_werte,
                      unsigned long& maxlen_types)
  {
    log(LOG_DEBUG,__FILE__,__LINE__,ERROR_UNSPECIFIED,
	"sercode(%lu) for %s called",pos,tname);
    char myaddr[9];
    char aktcode[9];
    char buf9[9];
    char buf33[33];

    myaddr[0]=0;
    known_class* ptr;
    serialize_elem* ptr1;

    ptr=known_classes;
    while (ptr)
      {
	if (!strcmp(tname,ptr->name))
	  {
	    strcpy(myaddr,ptr->addr);
	    break;
	  };
	ptr=ptr->next;
      };

    if (!myaddr[0])
      {
	snprintf(myaddr,9,"%.8lx",pos2);

	while (pos2+34>=maxlen_types) myrealloc(types,maxlen_types);
	strcpy(types+pos2,tname);
	pos2+=32;
	strcpy(types+pos2,"\n");
	pos2+=1;

	ptr1=elements;
	while (ptr1)
	  {
	    sercpy(buf33,ptr1->name);
	    while (pos2+34>=maxlen_types) myrealloc(types,maxlen_types);
	    strcpy(types+pos2,buf33);
	    pos2+=32;
	    strcpy(types+pos2,"\n");
	    pos2+=1;
	    ptr1=ptr1->next;
	  };
	
        while (pos2+34>=maxlen_types) myrealloc(types,maxlen_types);
	strcpy(types+pos2,"                                \n");
	pos2+=33;

	known_class* k=new known_class();
	if (!k) log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
	strcpy(k->name,tname);
	strcpy(k->addr,myaddr);
	k->next=known_classes;
	known_classes=k;
      };

    unsigned long elemcount=0;
    ptr1=elements;
    while (ptr1) { elemcount++; ptr1=ptr1->next; };
    unsigned long nextpos=pos+(elemcount+2)*10;
    while (nextpos+1>=maxlen_code) myrealloc(code,maxlen_code);
    code[nextpos]=0;

    strcpy(code+pos,"b");
    pos+=1;
    strcpy(code+pos,myaddr);
    pos+=8;
    strcpy(code+pos,"\n");
    pos+=1;

    ptr1=elements;
    while (ptr1)
      {
	char t[2];t[1]=0;t[0]=ptr1->type;
	log(LOG_DEBUG,__FILE__,__LINE__,ERROR_UNSPECIFIED,
	    "type %c, offset %lu, obj %lu for elem %s",
	    ptr1->type,ptr1->offset,obj,ptr1->name);
	switch (ptr1->type)
	  {
	  case SER_CLASSPTR:
	    {
	      char* ptr2;
	      if (ptr1->func)
		{
		  ptr2=*((char**) (*ptr1->func)(obj));
		}
	      else
		{
		  ptr2=*((char**) (obj+(ptr1->offset)));
		};
	      if (!ptr2)
		{
                  /* null pointer has position 0xffffffff */
		  snprintf(aktcode,10,"%.8x",0xffffffff);
		}
	      else
		{
		  snprintf(aktcode,10,"%.8lx",nextpos);
		  ptr1->desc->sercode(ptr2,code,nextpos,werte,pos1,types,
				      pos2,known_classes,maxlen_code,
				      maxlen_werte,maxlen_types);
		};
	      break;
	    };
	  case SER_CLASS:
	    {
	      char* ptr2;
	      if (ptr1->func)
		{
		  ptr2=(*ptr1->func)(obj);
		}
	      else
		{
		  ptr2=obj+(ptr1->offset);
		};
              if (!ptr2)
                {
                  /* null pointer has position 0xffffffff */
                  snprintf(aktcode,10,"%.8x",0xffffffff);
                }
              else
                {
		  snprintf(aktcode,10,"%.8lx",nextpos);
		  ptr1->desc->sercode(ptr2,code,nextpos,werte,pos1,types,pos2,
				      known_classes,maxlen_code,maxlen_werte,
				      maxlen_types);
		};
	      break;
	    };
	  case SER_STRING:
            {
	      char* wert;
	      if (ptr1->func)
		{
		  wert=(*ptr1->func)(obj);
		}
	      else
		{
		  wert=*((char**) (obj+(ptr1->offset)));
		};
	      if (wert)
		{
		  unsigned long len=strlen(wert);
		  snprintf(aktcode,10,"%.8lx",pos1);
		  snprintf(buf9,10,"%.8lx",len);
		  while (pos1+8+len+1>=maxlen_werte)
		    myrealloc(werte,maxlen_werte);
		  strcpy(werte+pos1,buf9);
		  pos1+=8;
		  strcpy(werte+pos1,wert);
		  pos1+=len;
		}
	      else
		{
		  /* null pointer has position 0xffffffff */
		  snprintf(aktcode,10,"%.8x",0xffffffff);
		};
              break;
            };
	  default:
	    {
	      long wert;
	      if (ptr1->func)
		{
		  wert=*((long*) (*ptr1->func)(obj));
		}
	      else
		{
		  wert=*((long*) (obj+(ptr1->offset)));
		};
	      snprintf(aktcode,10,"%.8lx",wert);
	      break;
	    };
	  };
	strcpy(code+pos,t);
	pos+=1;
	strcpy(code+pos,aktcode);
	pos+=8;
	strcpy(code+pos,"\n");
	pos+=1;

	ptr1=ptr1->next;
      };
    strncpy(code+pos,"effffffff\n",10);
    pos=nextpos;
    log(LOG_DEBUG,__FILE__,__LINE__,ERROR_UNSPECIFIED,
        "sercode() left");
    return;
  };

  inline void code(char*& code,char*& werte,char*& typen,unsigned long len,
		   char* ptr)
  {
    unsigned long pos=0;
    unsigned long pos1=0;
    unsigned long pos2=0;
    unsigned long maxlen_code=len;
    unsigned long maxlen_werte=len;
    unsigned long maxlen_typen=len;
    code=(char*) malloc(len);
    werte=(char*) malloc(len);
    typen=(char*) malloc(len);
    if (!code || !werte || !typen)
      log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
    known_class* k=0;
    sercode((char*) ptr,code,pos,werte,pos1,typen,pos2,k,
            maxlen_code,maxlen_werte,maxlen_typen);
    if (k) delete k;
    return;
  };
};

inline unsigned long gethex(char* src)
{
  unsigned long number;
  char buf[9];
  strncpy(buf,src,8);
  buf[8]=0;
  sscanf(buf,"%lx",&number);
  return number;
};

inline void serprint(char* c,char* t,char* w,unsigned long pos,unsigned long h)
{
  unsigned long i;
  unsigned long tpos;
  unsigned long aktelem=1;
  char elemname[33];
  char classname[33];
  if (!c[pos]) return;

  tpos=gethex(c+pos+1);
  pos+=10;

  for (i=0;i!=h;i++) printf(" ");printf("{\n");
  h=h+1;

  while (c[pos])
    {
      char type=c[pos];

      if (type==SER_ENDCLASS) break;
      
      for (i=0;i!=h;i++) printf(" ");

      unsigned long number=gethex(c+pos+1);

      elemname[0]=0;
      i=0;
      while (t[tpos+aktelem*33+i]!=' ' &&
             t[tpos+aktelem*33+i]!='\n')
        elemname[i]=t[tpos+aktelem*33+i];
        ++ i;
      elemname[i]=0;

      switch (type)
        {
        case SER_CLASS:
          {
            classname[0]=0;
            if (number!=0xffffffff)
              {
                unsigned long tpos1=gethex(c+number+1);
                i=0;
                while (t[tpos1+i]!=' ' &&
                       t[tpos1+i]!='\n')
                {
                    classname[i]=t[tpos1+i];
                    ++ i;
                }
                classname[i]=0;
              };
            if (!classname[0]) strcpy(classname,"unknown");

            if (number==0xffffffff)
              {
                printf("%s %s=NULL;\n",(char*) classname,(char*) elemname);
              }
            else
              {
                printf("%s %s=\n",(char*) classname,(char*) elemname);
                serprint(c,t,w,number,h);
              };
            break;
          };
        case SER_CLASSPTR:
          {
            classname[0]=0;
            if (number!=0xffffffff)
              {
                unsigned long tpos1=gethex(c+number+1);
                i=0;
                while (t[tpos1+i]!=' ' &&
                       t[tpos1+i]!='\n')
                { 
                    classname[i]=t[tpos1+i];
                    ++ i;
                }
                classname[i]=0;
              };
            if (!classname[0]) strcpy(classname,"unknown");

            if (number==0xffffffff)
              {
                printf("%s* %s=NULL;\n",(char*) classname,(char*) elemname);
              }
            else
              {
                printf("%s* %s=\n",(char*) classname,(char*) elemname);
                serprint(c,t,w,number,h);
              };
            break;
          };
        case SER_STRING:
          {
            if (number==0xffffffff)
              {
                printf("char* %s=NULL;\n",(char*) elemname);
              }
            else
              {
                unsigned long len=gethex(w+number);
                char* s=(char*) malloc(len+1);
		if (!s) log(LOG_FATAL,__FILE__,__LINE__,ERROR_OUT_OF_MEM,0);
                strncpy(s,w+number+8,len);
                s[len]=0;
                printf("char* %s=\"%s\";\n",(char*) elemname,s);
                free(s);
              };
            break;
          };
        case SER_LONG:
          {
            long value=(long) number;
            printf("long %s=%li;\n",(char*) elemname,value);
            break;
          };
        case SER_UNSIGNED_LONG:
          {
            printf("unsigned long %s=%lu;\n",elemname,number);
            break;
          };
        case SER_SHORT:
          {
            short value=(short) number;
            printf("short %s=%hi;\n",(char*) elemname,value);
            break;
          };
        case SER_UNSIGNED_SHORT:
          {
            unsigned short value=(unsigned short) number;
            printf("unsigned short %s=%hu;\n",(char*) elemname,value);
            break;
          };
        };
      pos+=10;
      aktelem++;
    };
  h=h-1;
  for (i=0;i!=h;i++) printf(" ");printf("};\n");
  return;
}
