#include "tparser.h"

#include <ctype.h>

namespace
{
    std::string XHide(std::string const &text)
    {
        std::vector<char> res;
            
        res.reserve(text.size() + 1);

        char c;
        char const *p = text.c_str();
        while ((c = *p++))
        {
            char rc;
            if ((c >= 'A' && c <= 'Z') ||
                (c >= '0' && c <= '9') ||
                 c == '' || c == '' || c == '' ||
                 c == '.' || c == ',')
            {
                rc = 'X';
            }
            else if (c == ' ' || c == '-')
            {
                rc = c;
            }
            else
            {
                rc = 'x';
            }
            res.push_back(rc);
        }
            
        res.push_back('\0');
        return &res[0];
    }

    inline double StringToDouble(char const *s)
    {
        return std::strtod(s, NULL);
    }
        
    inline long StringToLong(char const *s)
    {
        return std::strtol(s, NULL, 0);
    }

    void StringToLower(std::string &s)
    {
        size_t size = s.size();
        for (size_t n = 0; n < size; ++n)
        {
            s[n] = static_cast<char>(tolower(s[n]));
        }
    }

    bool CommandEnthaelt(std::string const &s1, std::string const &s2)
    {
        return s1.find(s2, 0) != std::string::npos;
    }
    
    bool CommandUenthaelt(std::string const &s1, std::string const &s2)
    {
        std::string s1Copy = s1;
        std::string s2Copy = s2;
        StringToLower(s1Copy);
        StringToLower(s2Copy);
        return CommandEnthaelt(s1Copy, s2Copy);
    }
    
    bool CommandUenthaeltnicht(std::string const &s1, std::string const &s2)
    {
        return !CommandUenthaelt(s1, s2);
    }
    
    bool CommandRenthaelt(std::string const &s1, std::string const &s2)
    {
        return CommandEnthaelt(s2, s1);
    }
    
    bool CommandEnthaeltnicht(std::string const &s1, std::string const &s2)
    {
        return !CommandEnthaelt(s1, s2);
    }
    
    bool CommandIsinset(std::string const &s1, std::string const &s2)
    {
        size_t pFrom = 0, pTo;

        char const *s1Str = s1.c_str();
        char const *s2Str = s2.c_str();

        std::map<std::string, bool> s1Tokens;

        while ((pTo = s1.find(',', pFrom)) != std::string::npos)
        {
            s1Tokens[std::string(s1Str + pFrom, s1Str + pTo)] = true;
            pFrom = pTo + 1;
        }
        s1Tokens[s1Str + pFrom] = true; // last token

        pFrom = 0;
        while ((pTo = s2.find(',', pFrom)) != std::string::npos)
        {
            if (s1Tokens[std::string(s2Str + pFrom, s2Str + pTo)])
            {
                return true;
            }
            pFrom = pTo + 1;
        }
        if (s1Tokens[s2Str + pFrom]) // last token
        {
            return true;
        }
        return false;
    }

    bool CommandIsnotinset(std::string const &s1, std::string const &s2)
    {
        return !CommandIsinset(s1, s2);
    }

    bool CommandGleich(std::string const &s1, std::string const &s2)
    {
        return s1 == s2;
    }
    
    bool CommandLeer(std::string const &s)
    {
        return s.empty();
    }
    
    bool CommandNichtleer(std::string const &s)
    {
        return !s.empty();
    }
    
    bool CommandNichtgleich(std::string const &s1, std::string const &s2)
    {
        return s1 != s2;
    }

    bool CommandNumericEqual(std::string const &s1, std::string const &s2)
    {
        return StringToDouble(s1.c_str()) == StringToDouble(s2.c_str());
    }

    bool CommandNumericNotequal(std::string const &s1, std::string const &s2)
    {
        return !CommandNumericEqual(s1, s2);
    }

    bool CommandNumericLess(std::string const &s1, std::string const &s2)
    {
        return StringToDouble(s1.c_str()) < StringToDouble(s2.c_str());
    }

    bool CommandNumericLessOrEqual(std::string const &s1, std::string const &s2)
    {
        return StringToDouble(s1.c_str()) <= StringToDouble(s2.c_str());
    }

    bool CommandNumericGreater(std::string const &s1, std::string const &s2)
    {
        return StringToDouble(s1.c_str()) > StringToDouble(s2.c_str());
    }

    bool CommandNumericGreaterOrEqual(std::string const &s1, std::string const &s2)
    {
        return StringToDouble(s1.c_str()) >= StringToDouble(s2.c_str());
    }

    bool CommandIsodd(std::string const &s)
    {
        return StringToLong(s.c_str()) & 1;
    }

    bool CommandIseven(std::string const &s)
    {
        return !CommandIsodd(s);
    }

    bool CommandAnfanggleich(std::string const &s1, std::string const &s2)
    {
        return s1.size() <= s2.size() && s1 == s2.substr(0, s1.size());
    }

    bool CommandEndegleich(std::string const &s1, std::string const &s2)
    {
        return s1.size() <= s2.size() && s1 == s2.substr(s2.size() - s1.size());
    }

    bool CommandDa(std::string const &s)
    {
        return !s.empty(); // Note: See special handling in ParseCommand!
    }

    bool CommandNichtda(std::string const &s)
    {
        return s.empty(); // Note: See special handling in ParseCommand!
    }
}

TemplateParser::CommandMap TemplateParser::cmds_;

inline type_value &TemplateParser::LookupVarValue(char const *var, bool &found) {
    return lookupFnc_(var, found, lookupEnv1_, lookupEnv2_);
}

inline type_value TemplateParser::ExecFunc(FunctionExpr &funcExpr) {
    flowctrl call = execFnc_(funcExpr.functionName.c_str(),
                             lookupEnv1_, lookupEnv2_,
                             funcExpr.args);
    if (call.ctrl == FLOW_OK) {
        return call.returnval;
    }
    
    dest_ = ::string(srcBegin_,  srcPtr_);
    Write("<br>Template fehlerhaft\n");
    error_msg_ = "Template fehlerhaft";
    error_occured_ = true;
    flow_error_occured_ = true;
    throw call;
}

char const *TemplateParser::LookupVar(char const *var, bool &found)
{
    type_value &tokens = LookupVarValue(var, found);
    static char ret[120];
    switch (tokens.type)
    {
    case TYPE_VALUE_STRING:
        if (!defined(tokens.stringval))
        {
            return "";
        }
        return tokens.stringval.c_str();
    case TYPE_VALUE_LONG:
        ::snprintf(ret,80,"%li", tokens.val.longval);
        return ret;
    case TYPE_VALUE_ULONG:
        ::snprintf(ret,80,"%lu", tokens.val.ulongval);
        return ret;
    case TYPE_VALUE_DOUBLE:
        ::snprintf(ret,80,"%f", tokens.val.doubleval);
        return ret;
    default:
        return "";
    }
}

TemplateParser::TemplateParser(TemplateParserLookupVarFnc lookupFnc,
                               TemplateParserExecFuncFnc execFnc,
                               void *lookupEnv1, void *lookupEnv2)
:   lookupFnc_(lookupFnc), execFnc_(execFnc),
    lookupEnv1_(lookupEnv1), lookupEnv2_(lookupEnv2)
{
    if (cmds_.empty())
    {
        cmds_["enthaelt"] = Command(CommandEnthaelt);
        cmds_["renthaelt"] = Command(CommandRenthaelt);
        cmds_["enthaeltnicht"] = Command(CommandEnthaeltnicht);
        cmds_["uenthaelt"] = Command(CommandUenthaelt);
        cmds_["uenthaeltnicht"] = Command(CommandUenthaeltnicht);
        cmds_["leer"] = Command(CommandLeer);
        cmds_["nichtleer"] = Command(CommandNichtleer);

        cmds_["isinset"] = Command(CommandIsinset);
        cmds_["isnotinset"] = Command(CommandIsnotinset);

        cmds_["gleich"] = Command(CommandGleich);
        cmds_["nichtgleich"] = Command(CommandNichtgleich);

        cmds_["="] = Command(CommandNumericEqual);
        cmds_["!="] = Command(CommandNumericNotequal);
        cmds_["<"] = Command(CommandNumericLess);
        cmds_["<="] = Command(CommandNumericLessOrEqual);
        cmds_[">"] = Command(CommandNumericGreater);
        cmds_[">="] = Command(CommandNumericGreaterOrEqual);
        cmds_["iseven"] = Command(CommandIseven);
        cmds_["isodd"] = Command(CommandIsodd);

        cmds_["anfanggleich"] = Command(CommandAnfanggleich);
        cmds_["endegleich"] = Command(CommandEndegleich);

        cmds_["da"] = Command(CommandDa);
        cmds_["nichtda"] = Command(CommandNichtda);
    }
}

bool TemplateParser::BraceAhead()
{
    return srcEnd_ - srcPtr_ >= 3 &&
           srcPtr_[0] == '|' &&
           (srcPtr_[1] == '{' || srcPtr_[1] == '}') &&
           srcPtr_[2] == '|';
}

bool TemplateParser::ElseAhead()
{
    if (srcEnd_ - srcPtr_ >= 5 &&
        srcPtr_[0] == 'e' &&
        srcPtr_[1] == 'l' &&
        srcPtr_[2] == 's' &&
        srcPtr_[3] == 'e' &&
        srcPtr_[4] == '|')
    {
        srcPtr_ += 5;
        return true;
    }
    return false;
}

void TemplateParser::Error(char const *msg)
{
    dest_ = ::string(srcBegin_,  srcPtr_);
    Write("<br>Template fehlerhaft: ");
    Write(msg);
    Write("\n");
    error_msg_ = "Template fehlerhaft: ";
    error_msg_ += msg;
    error_occured_ = true;    
    throw 1;
}

char const *TemplateParser::FindNextPipe(char const *p)
{
    while (p != srcEnd_ && *p != '|')
    {
        ++p;
    }
    return p;
}

char const *TemplateParser::FindNextPipeOrParen(char const *p)
{
    while (p != srcEnd_ && *p != '|' && *p != ')')
    {
        ++p;
    }
    return p;
}

char const *TemplateParser::FindNextPipeOrSharp(char const *p)
{
    while (p != srcEnd_ && *p != '|' && *p != '#')
    {
        ++p;
    }
    return p;
}

Command const &TemplateParser::GetCommand(std::string const &cmdName)
{
    CommandIterator it = cmds_.find(cmdName);
    if (it == cmds_.end())
    {
        Error("Unbekannter Befehl.");
    }
    return it->second;
}

::string TemplateParser::GetResult(bool &error_occured,
                                   bool &flow_error_occured,
                                   std::string &error_msg)
{
    ::string ret = dest_;
    dest_ = 0;
    error_occured = error_occured_;
    flow_error_occured = flow_error_occured_;
    error_msg = error_msg_;
    return ret;
}

flowctrl TemplateParser::GetFlowError() {
    return flowError_;
}

void TemplateParser::Parse()
{
    dest_ = "";

    srcBegin_ = src_.begin();
    srcEnd_ = src_.end();
    srcPtr_ = srcBegin_;
    depth_ = 0;
    error_msg_ = "";
    error_occured_ = false;
        flow_error_occured_ = false;

    try
    {
        ParseRange(false);
    }
    catch (flowctrl call) {
        flowError_ = call;
        flow_error_occured_ = true;
    }
    catch (int e)
    {
    }
}

void TemplateParser::ParseBody(bool skip)
{
    for (;;)
    {
        ParseText(skip);
        if (BraceAhead())
        {
            if (srcPtr_[1] == '{')
            {
                srcPtr_ += 3;
                depth_ += 1;
                ParseRange(skip);
                // Brace ahead!
            }
            else // '}'
            {
                Error("Unerwartetes Blockende.");
            }
        }
        else
        {
            if (Eof())
            {
                Error("| erwartet.");
            }
            srcPtr_ += 1;
            break;
        }
    }
}

void TemplateParser::ParseCommand(bool skip)
{
    if (Eof() || *srcPtr_ != '|')
    {
        Error("Befehl erwartet.");
    }
    ++ srcPtr_; // skip |
    bool value = ParseORExpression(skip);
    SkipWhite();
    if (Eof() || *srcPtr_ != '|')
    {
        Error("Pipe '|' erwartet.");
    }
    ++srcPtr_;
    if (skip)
    {
        ParseBody(true);
        ParseElse(true);
    }
    else
    {
        ParseBody(!value);
        ParseElse(value);
    }
}

bool TemplateParser::ParseORExpression(bool skip)
{
    bool ret = ParseANDExpression(skip);
    while (ParseOperator("OR"))
    {
        // ParseANDExpression has side effects!
        bool v = ParseANDExpression(skip); 
        ret = ret || v;
    }
    return ret;
}

bool TemplateParser::ParseANDExpression(bool skip)
{
    bool ret = ParseNOTExpression(skip);
    while (ParseOperator("AND"))
    {
        // ParseNOTExpression has side effects!
        bool v = ParseNOTExpression(skip); 
        ret = ret && v;
    }
    return ret;
}

bool TemplateParser::ParseNOTExpression(bool skip)
{
    bool notOp = false;
    while (ParseOperator("NOT"))
    {
        notOp = !notOp;
    }
    bool ret = ParseSimpleExpression(skip); 
    if (notOp)
        ret = !ret;
    return ret;
}

bool TemplateParser::ParseOperator(char const *op)
{
    SkipWhite();
    char const *src = srcPtr_;
    while (*op)
    {
        if (src == srcEnd_ || *src++ != *op++)
            return false;
    }
    srcPtr_ = src;
    return true;
}

void TemplateParser::SkipWhite()
{
    while (srcPtr_ != srcEnd_ &&
           (*srcPtr_ == ' '
            or *srcPtr_ == '\n'
            or *srcPtr_ == '\r'
            or *srcPtr_ == '\t'))
    {
        ++srcPtr_;
    }
}

bool TemplateParser::ParseSimpleExpression(bool skip)
{
    SkipWhite();
    if (Eof())
    {
        Error("Logischer Ausdruck erwartet.");
    }

    if (*srcPtr_ == '(')
    {
        // Parse parenthised expression.
        ++srcPtr_;
        bool ret = ParseORExpression(skip);
        SkipWhite();
        if (Eof() || *srcPtr_ != ')')
        {
            Error("Schliessende Klammer ')' erwartet.");
        }
        ++srcPtr_;
        return ret;
    }

    // Simple expression.
    std::string cmdName = ParseCommandName();
    Command const &cmd = GetCommand(cmdName);
    bool found;
    std::string arg1 = ParseFirstCommandArgument(skip, found);
    if (!skip)
    {
        if (cmd == CommandDa || cmd == CommandNichtda)
        {
            arg1 = found ? "+" : "";
        }
        else if (!found)
        {
            Error("Variable nicht vorhanden.");
        }
    }
    bool value = false;
    if (cmd.ArgCount() == 2)
    {
        ++srcPtr_;
        std::string arg2 = ParseSecondCommandArgument(skip);
        if (!skip)
        {
            value = cmd.Execute(arg1, arg2);
        }
    }
    else if (!skip)
    {
        value = cmd.Execute(arg1);
    }
    return value;
}

std::string TemplateParser::ParseCommandName()
{
    char const *ptrPipe = FindNextPipe(srcPtr_);
    if (ptrPipe == srcEnd_)
    {
        Error("Befehlsname nicht abgeschlossen.");
    }
    std::string ret(srcPtr_, ptrPipe);
    srcPtr_ = ptrPipe + 1; // skip the |
    return ret;
}

bool TemplateParser::ParseDoubleSharp()
{
    if (srcEnd_ - srcPtr_ >= 2 &&
        *srcPtr_ == '#' &&
         srcPtr_[1] == '#')
    {
        srcPtr_ += 2;
        return true;
    }
    return false;
}

void TemplateParser::ParseElse(bool skip)
{
    if (ElseAhead())
    {
        ParseBody(skip);
    }
}

std::string TemplateParser::ParseFirstCommandArgument(bool skip, bool &found)
{
    char const *ptrPipe = FindNextPipeOrParen(srcPtr_);
    if (ptrPipe == srcEnd_)
    {
        Error("Erstes Befehlsargument nicht abgeschlossen.");
    }
    std::string var("a_");
    var += std::string(srcPtr_, ptrPipe);
    std::string lookupVal = LookupVar(var.c_str(), found);
    srcPtr_ = ptrPipe;
    return lookupVal;
}

std::string TemplateParser::ParseSecondCommandArgument(bool skip)
{
    char const *ptrPipe = FindNextPipeOrParen(srcPtr_);
    if (ptrPipe == srcEnd_)
    {
        Error("Zweites Befehlsargument nicht abgeschlossen.");
    }
    std::string var("a_");
    var += std::string(srcPtr_, ptrPipe);
    bool found;
    std::string lookupVal = LookupVar(var.c_str(), found);
    if (!skip && !found)
    {
        lookupVal = std::string(srcPtr_, ptrPipe);
    }
    srcPtr_ = ptrPipe;
    return lookupVal;
}

void TemplateParser::ParseLiteralText(bool skip) // May be empty, if eof.
{
    char const *ptrPipeSharp = FindNextPipeOrSharp(srcPtr_);
    
    if (!skip)
    {
        Write(srcPtr_, ptrPipeSharp);
    }
    srcPtr_ = ptrPipeSharp;
}

void TemplateParser::ParseRange(bool skip)
{
    for (;;)
    {
        ParseText(skip);
        if (Eof())
        {
            break;
        }
        // | ahead
        if (BraceAhead())
        {
            if (srcPtr_[1] == '{')
            {
                srcPtr_ += 3;
                depth_ += 1;
                ParseRange(skip);
                continue;
            }
            else // |}|
            {
                if (depth_ == 0)
                {
                    Error("Blockende ohne Blockanfang.");
                }
                srcPtr_ += 3;
                depth_ -= 1;
                return;
            }
        }
        ParseCommand(skip);
    }
    ParseText(skip);
    if (depth_)
    {
        Error("Blockende erwartet.");
    }
}

type_value TemplateParser::ExecuteFunctionExpr(CharRange &expr,
                                               bool &found) {
    SkipWs(expr);
    type_value literalValue;
    if (ParseLiteral(expr, literalValue)) {
        return literalValue;
    }
    CharRange const name = ParseName(expr);
    expr.begin = name.end;
    SkipWs(expr);
    if (!expr.empty() && expr[0] == '(') {
        found = true;
        ++ expr.begin;
        SkipWs(expr);
        FunctionExpr fexpr(::string(name.begin, name.end));
        if (!expr.empty() &&  expr[0] != ')') {
            ParseFunctionArgs(expr, fexpr);
            SkipWs(expr);
        }
        if (expr.empty() || expr[0] != ')') {
            Error(expr, "Schliessende Klammer ')' erwartet.");
        }
        ++ expr.begin;
        SkipWs(expr);
        return ExecFunc(fexpr);
    } else {
        bool f;
        type_value ret = LookupVarValue((std::string("a_") + std::string(name.begin, name.end)).c_str(), f);
        if (!f) {
            ret.type = TYPE_VALUE_STRING;
            ret.stringval = string(name.begin, name.end);
        }
        found = found || f;
        return ret;
    }
}

std::string TemplateParser::ExecuteFunctionExprToStr(CharRange const &expr,
                                                     bool &found) {
    found = false;
    CharRange ex = expr;
    SkipWs(ex);
    type_value value = ExecuteFunctionExpr(ex, found);
    if (!ex.empty()) {
        Error(ex, "Syntax Error."); 
    }
    char ret[128];
    switch (value.type) {
    case TYPE_VALUE_STRING:
        if (!defined(value.stringval))
        {
            return "";
        }
        return value.stringval.c_str();
    case TYPE_VALUE_LONG:
        ::snprintf(ret, 80, "%li", value.val.longval);
        return ret;
    case TYPE_VALUE_ULONG:
        ::snprintf(ret, 80, "%lu", value.val.ulongval);
        return ret;
    case TYPE_VALUE_DOUBLE:
        ::snprintf(ret, 80, "%f", value.val.doubleval);
        return ret;
    default:
        return "";
    }
}

void TemplateParser::ParseReplaceExpr(bool skip)
{
    char const *p = srcPtr_;
    while (p != srcEnd_)
    {
        if (*p == '#' &&
            p + 1 != srcEnd_ &&
            p[1] == '#')
        {
            if (!skip)
            {
                CharRange expr(0, 0);
                // XX behandeln.
                bool hide;
                if (p - srcPtr_ >= 2 && p[-1] == 'X' && p[-2] == 'X')
                {
                    expr.begin = srcPtr_;
                    expr.end = p - 2;
                    hide = true;
                }
                else
                {
                    expr.begin = srcPtr_;
                    expr.end = p;
                    hide = false;
                }
                bool found;
                std::string value = ExecuteFunctionExprToStr(expr, found);
                if (!found)
                {
                    value = std::string(srcPtr_ - 2, p + 2);
                }
                else if (hide)
                {
                    value = XHide(value);
                }
                Write(value.c_str());
            }
            srcPtr_ = p + 2;
            return;
        }
        ++ p;
    }
    Error("## erwartet.");
}

/*void TemplateParser::ParseReplaceVar(bool skip)
{
    char const *p = srcPtr_;
    while (p != srcEnd_)
    {
        if (*p == '#' &&
            p + 1 != srcEnd_ &&
            p[1] == '#')
        {
            if (!skip)
            {
                std::string var = "a_";
                // XX behandeln.
                bool hide;
                if (p - srcPtr_ >= 2 && p[-1] == 'X' && p[-2] == 'X')
                {
                    var += std::string(srcPtr_, p - 2);
                    hide = true;
                }
                else
                {
                    var += std::string(srcPtr_, p);
                    hide = false;
                }
                bool found;
                std::string value = LookupVar(var.c_str(), found);
                if (!found)
                {
                    value = std::string(srcPtr_ - 2, p + 2);
                }
                else if (hide)
                {
                    value = XHide(value);
                }
                Write(value.c_str());
            }
            srcPtr_ = p + 2;
            return;
        }
        ++ p;
    }
    Error("## erwartet.");
}*/

void TemplateParser::ParseText(bool skip) // May be empty, if eof.
{
    for (;;)
    {
        ParseLiteralText(skip);
        if (Eof() || *srcPtr_ == '|')
        {
            return;
        }
        if (ParseDoubleSharp())
        {
            ParseReplaceExpr(skip);
        }
        else // *srcPtr_ == '#'
        {
            if (!skip)
            {
                Write("#");
            }
            ++ srcPtr_;
        }
    }
}

void TemplateParser::SetSrc(const ::string &src)
{
    src_ = src;
}

void TemplateParser::Write(char const *strBegin, char const *strEnd)
{
    Write(strBegin, strEnd - strBegin);
}

void TemplateParser::Write(char const *str)
{
    Write(str, std::strlen(str));
}

inline bool IsNameChar(char c) {
    return c >= 'a' && c <= 'z'
           || c >= 'A' && c <= 'Z'
           || c >= '0' && c <= '9'
           || c == '_';
}

TemplateParser::CharRange TemplateParser::ParseName(TemplateParser::CharRange &s) {
    char const *start = s.begin;
    if (s.empty() || !IsNameChar(*s.begin)) {
        Error(s, "Identifikator erwartet.");
    }
    do {
        ++ s.begin;
    } while (!s.empty() && IsNameChar(*s.begin));
    CharRange ret(start, s.begin);
    return ret;
}

void TemplateParser::SkipWs(CharRange &s) {
    for (;;) {
        if (s.empty()) {
            return;
        }
        char const c = s[0];
        switch (c) {
        case '\t':
        case '\n':
        case '\r':
        case ' ':
            ++ s.begin;
            break;
        default:
            return;
        }
    }
}

void TemplateParser::ParseFunctionArg(CharRange &args, FunctionExpr &fexpr) {
    bool foundDummy;
    type_value arg = ExecuteFunctionExpr(args, foundDummy);
    fexpr.addArg(arg);
}

void TemplateParser::ParseFunctionArgs(CharRange &args, FunctionExpr &fexpr) {
    ParseFunctionArg(args, fexpr);
    for (;;) {
        SkipWs(args);
        if (args.empty() || args[0] != ',') {
            break;
        }
        ++ args.begin;
        SkipWs(args);
        ParseFunctionArg(args, fexpr);
    }
}

bool TemplateParser::ParseLiteral(CharRange &expr, type_value &literalValue) {
    if (ParseStringLiteral(expr, literalValue)) {
        return true;
    }
    CharRange ex = expr;
    type_value dv;
    bool const longParsed = ParseLongLiteral(expr, literalValue);
    bool const doubleParsed = ParseDoubleLiteral(ex, dv);
    if (longParsed) {
        if (doubleParsed && ex.begin > expr.begin) {
            expr = ex;
            literalValue = dv;
        }
        return true;
    }
    if (doubleParsed) {
        expr = ex;
        literalValue = dv;
        return true;
    }
    return false;
}

int TemplateParser::HexValue(CharRange const &expr, char c) {
    if (c >= '0' && c <= '9') {
        return c - '0';
    } else if (c >= 'A' && c <= 'F') {
        return c - 'A' + 10;
    } else if (c >= 'a' && c <= 'f') {
        return c - 'a' + 10;
    } else {
        Error(expr, "Ungueltige Escapesequenz."); // Schmeisst Ausnahme
        return -1; // Compiler ruhigstellen.
    }
}

bool TemplateParser::ParseStringLiteral(CharRange &expr, type_value &literalValue) {
    if (expr.empty() || expr[0] != '"') {
        return false;
    }
    ++ expr.begin;
    std::string s = "";
    while (!expr.empty() && expr[0] != '"') {
        char c;
        if (expr[0] == '\\') {
            ++ expr.begin;
            if (expr.empty()) {
                Error(expr, "Ungueltige Escapesequenz.");
            }
            char const esc = expr[0];
            ++ expr.begin;
            switch (esc) {
            case 'n':
                c = '\n';
                break;
            case 't':
                c = '\t';
                break;
            case 'v':
                c = '\v';
                break;
            case 'b':
                c = '\b';
                break;
            case 'r':
                c = '\r';
                break;
            case 'f':
                c = '\f';
                break;
            case 'a':
                c = '\a';
                break;
            case '\"':
                c = '\"';
                break;
            case '\\':
                c = '\\';
                break;
            case 'x':
                if (expr.size() < 2) {
                    Error(expr, "Ungueltige Escapesequenz.");
                }
                c = 16 * HexValue(expr, expr[0]) + HexValue(expr, expr[1]);
                expr.begin += 2;
                if (c == 0) {
                    Error(expr, "Ungueltige Escapesequenz.");
                }
                break;
            default:
                Error(expr, "Ungueltige Escapesequenz.");
            }
        } else {
            c = *expr.begin++;
        }
        s += c;
    }
    if (expr.empty()) {
        Error(expr, "Stringliteral nicht abgeschlossen.");
    }
    ++expr.begin;
    literalValue.type = TYPE_VALUE_STRING;
    literalValue.stringval = s.c_str();
    return true;
}

bool TemplateParser::ParseLongLiteral(CharRange &expr, type_value &literalValue) {
    CharRange ex = expr;
    if (ex.empty()) {
        return false;
    }
    bool negative = false;
    if (ex[0] == '-') {
        ++ex.begin;
        if (ex.empty()) {
            return false;
        }
        negative = true;
    }
    char c = *ex.begin++;
    if (c < '0' || c > '9') {
        return false;
    }
    long value = c - '0';
    while (!ex.empty() && ex[0] >= '0' && ex[0] <= '9') {
        c = *ex.begin++;
        value = value * 10 + c - '0';
    }
    if (negative) {
        value = -value;
    }
    literalValue.type = TYPE_VALUE_LONG;
    literalValue.val.longval = value;
    expr = ex;
    return true;
}

bool TemplateParser::ParseDoubleLiteral(CharRange &expr, type_value &literalValue) {
    CharRange ex = expr;
    if (ex.empty() || (ex[0] != '-' && ex[0] != '.' && (ex[0] < '0' || ex[0] > '9'))) {
        return false;
    }
    errno = 0;
    std::string const s(expr.begin, expr.end);
    char const *sv = s.c_str();
    char *endPtr;
    double value = strtod(sv, &endPtr);
    if (errno) {
        return false;
    }
    ex.begin += endPtr - sv;
    literalValue.type = TYPE_VALUE_DOUBLE;
    literalValue.val.doubleval = value;
    expr = ex;
    return true;
}
